/* objects and programs for the transmitter computational function:
  - read control functions
  - compute several flight control variables
  - transmit a signal for the servos

  ----Classes------------------------------
  ---Calc = main object for the current model
    list<Operation> operations
    vector<Var>vars one single vector with all the variables (index 0 -1000 maximum)
    vector<string>libs , for some strings
    vector<string>errors

  ---Operation  each operation is the translation of one source line
        = code operation
        + vector<short> with the index of each variable (0-1 outvar , 0-n invar)

  ---Var = caracteristics of one variable
    typeOrig= origine, reset...
    typeNum = numeric;, logical or string variable
    short value ( numerical, logical, pointer to another index)


*/
/*============================================================================
 Copyright (C) 2008 -2012 Robert Spilleboudt

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
============================================================================*/
#ifndef CALC_H
#define CALC_H
#include <string>
#include <vector>
#include<queue>

enum VarNum {NUM , LOGICAL};    // defines the format of the variable
enum VarOrig { IN , OUT, PARAM_VVAR , WORK_MEM, WORK , PARAM_PVAR};
/*  in = hardware in,initialized before each loop
    out=hw out
    param_pvar=       pvar: INIT vvar condition
    work_mem initialized before the firs loop
    work initialized before each loop
*/
#define TOINT 100  //conversion to internal format integer short: a variable 50.75 has a internal value 5075
#define TOVAR 20000
#define NOT_AVAILABLE -12500    //this value is not available
/*cf operation:
argument < TOVAR is a constant (example 575, divided by TOINT, is the constant 5.75
argument >TOVAR, example 20015, subtact TOVAR-> 15 =index to the variable 15*/
bool convert (string const& s, short& y , short divider =0); //optionnal divider example 3 = divide by 1000
short valcheck(short x); // the internal variable x constrained to limits
string lcdTimer(int timer) ; //timer units= 1/100 minute , returns a string formatted as mm:ss  or h.mm-> calcrun.cpp
class Var {
public:
    string name;
    //VarOrig typeOrig;
		VarOrig getTypeOrig();
		void putTypeOrig(VarOrig);
    VarNum  typeNum ;
    short value;
public:
    string valueS(int f); // f default = 0
    //the calling program knows the variable type
 //   short getNum();
 //   bool  getBool();
 //   string getString();
    Var(string name, VarOrig t, VarNum u );
    string print();
private:
	VarOrig typeOrig;
} ;

class Operation { //operation = 1 line of the source program !!only during the source file decoding
public:
    char code; //code identifying the operation
    // for the error messages: reference to the line number of the source file
    short linenumber; //program line in the base mdx file
    short linenumberM; //program line in the module mdx file, or 0
    vector<short> vars; // >=20000  value-20000 =index to a variable
    // <20000  constant value

    Operation();
    Operation( int n,int linenumber, int linenumbermodule, char code);
    void debugPrint(void) ;
};
class Dtrim { //one object for each hardware digital trim (2 buttons)
private:
    int timer; //counting 10sec ticks
		int timerSnd; //last sound was issued at thie getTicker

public:
    Dtrim() ; //construction of a trim -without asigned params
    void get (int dirp ,int dirm); //process every 10 msec the buttons and update the correponding parameter
    short indexPvarNormalTrim; //index of the normal PVAR (level 0 or 1, assigned by the rccl DTRIMx instruction)
    short indexPvarTuneParam; //index of the inflight PVAR (level 2, assigned by the parameters update menu)
    //index not assigned is  = -1
};
class Calc {
private:
    vector<Var> calc_vars; // access to the variables with a index from 0 to...
    vector<string>calc_lib; //alarm messages and other string variables
    queue<string>logq; // logged strings 
    int freeRam; //free ram after the load of this model   
	int currentBreak; // current break message from this iop
    int currentLog; // iop of the current log or -1 (to log a header record )
    int currentLogTime; //GetTime for the former log operation

public:
    vector<string>errors; //error messages
    Dtrim dtrimArray[4];
    short dtrim_level;  // 3 position switch
    /* 0 = normal       Autotrim =disabled  digital trims = normal trim of each stick
       1 = fast trim    Autotrim =enabled   digital trims = normal trim of each stick
       2 = param trim   Autotrim =enabled   digital trims assigned to parameters
    */
    string lcd[8]; //lines for the in-flight display (-> menu_normal)
		int lcd_protect[8]; // until this tick value the displayed lcd [i] may not be updated/erased
    int iop,iopend;

    string modelname ; //model loaded in this Calc object
    string phasename ;
    bool ready_to_calc ;//true if the model is ready and the calcrun may be called (if errors.length()>0 these messages are warnings )
    //progressive transition after a phse change
    int slow_done; //elapsed ticker cycles
    int slow_end;  // duration of the transition
    bool slow_start;
    vector<string> error ;//a vector with 0-n error messages (none is duplicated)
	int rx_count_logged;// telemetry count for the last logged record
    short addVar(string& name, int lineNumber,int lneNumberM, VarNum typeNum , VarOrig typeOrig);
    short getVarIndex(string& name);//this is a alphabetical search of a var (not a pvar)
    short getVarIndexP(string& name);//this is a alphabetical search of a pvar
    Var   getVar(short i);// get the variable indexed i(vvar) or NOT_AVAILABLE
    Var *  getVarptr(short i);// get the pointer  to the variable indexed i
	int getVarP( string& name); //get the value of a parameter 
    void  setVarN(string name, int value); // set the numerical value of a var defined by the name 
    void  setVarL(string name, bool value); // set the logical value of a var defined by the name 
    int   getnVar(void); //number of variables
    //----- operations, with a specific storage----------------------------
    bool  putOperation(Operation op); //store a operation, return false if error
    void  setOperationBegin() ; //set the operation counter at the beginning
    void  setOperationNext(); //set the operation counter to the next operation (may be  a EOF)
		int  getOperationIndex();//get the index of the current operation
    void  prtOperation(string z);//print the current operation
    //some functions to isolate the calc logic from the storage choice
    char getOp_code();//get the operation code
    int  getOp_linenumber();//linenumber for error processing
    int  getOp_linenumberM();//linenumber within a module
    int  getOp_len(); //returns the number of variables
    //the following indexes i are the index of the variable in the operation line (0...n)
    Var * getOp_Varptr(int i); //get a pointer to the variable at index i in the operation
    int  getOp_valueN(int i); // get the value of the variable indexed i (numerical )-with indirect acces PVAR
    bool getOp_valueL(int i); // get the value of the variable indexed i (logical )
    string getOp_valueS(int i);//return the value of the string variable at position i in this operation
    void setOp_valueN(int i,int value ); // set the value of the variable indexed i (numerical )
    void setOp_valueL(int i,bool value); // set the value of the variable indexed i (logical )

    Calc( string filename);//loads the model filename format 8 char without extension
		~Calc(); //destructor
    void calcloop();
    void erreur(int linenumber,int linemodulenumber, string message);
    bool param_restore(); // read the parameters file filename , ,without the extension
    bool param_save(); // write the parameters file for this model and the global parameters file 
    void varDump(int select);//select=0 only vars select=1 vars+strings + operations
    void run();
	void kick(); //reset the autocutoff ticker
	void logger_put(string s); //put a string in the queue to the lcd sd card
	void logger_sd(); // get a string from the queue and send this to the lcd
	
	int autocutoff_ticker; // future ticker value for a automatic power-off
	private:
	bool expandfile(string filename);//expands one file (insert the CALL /CALLP replacing the parameters)
	void insert (vector<string>& x, string& y,int line, int modline) ; //insert one text line into the calc data
	int autocutoff_sticks; // current value of sum of the sticks
    // hw depending
public:
    void hwinit(); // load eaxh hardware variable (once, in the constructor of the Calc object)
    void hwread(); // once each rcloop,read the inputs from the microprocessor
    void hwout(); // once each rcloop, send the output to the transmission system (PPM or other)

// specific for validate
  void Validate();
  void Valid_err (const string err); //error message
  bool Valid_len (int n);//return true when error
	bool Valid_typ (int n, VarNum t);
	bool Valid_orig (int n, VarOrig t);
}  ;

class Curve {// main component of a mixer  -> mixer.cpp
private:
    vector<short> vx,vy;
public:
    Curve();
    void addPoint(short x, short y);
    short get(short x);

};

int interpol(int x,int x1,int y1,int x2,int y2);

struct code_operation {// code operation definition
public:
    string name;
    char codeop;
    bool outvar;
    VarNum  typenum;
    bool statevar; //true = reset only before the first loop
    int  intvar;//internal variables ( for instance: time depending operations)
	  //int  invar;//fixed number of input var of -1 if not controlled
};
bool get_code_operation(code_operation& c, string name);
//--------------------- definition of the code operation---------------------------------
#define	 OP_IF             '\x01'
#define	 OP_ENDIF	    '\x02'
#define	 OP_PHASE   	'\x03'
#define	 OP_SLOW	      '\x04'
#define	 OP_INIT	       '\x05'
#define	 OP_ADD	       '\x06'
#define	 OP_SUB	       '\x07'
#define	 OP_MULT     	'\x08'
#define	 OP_MULT_PC '\x09'
#define	 OP_EQU	      '\x10'
#define	 OP_EQU_PARAM	'\x11'
#define	 OP_LINE	       '\x12'
#define	 OP_CURVE	  '\x13'
#define	 OP_EXPO	   '\x14'
#define	  OP_OUT	   '\x15'
#define	  OP_OUTN	 '\x16'
#define	  OP_SERV	 '\x17'
#define	  OP_ADDS	 '\x18'
#define	  OP_SUBS	 '\x19'
#define	  OP_ADDAS	'\x20'
#define	  OP_SUBAS	'\x21'
#define	  OP_AND	    '\x22'
#define	  OP_OR	     '\x23'
#define	  OP_NOT  	 '\x24'
#define	  OP_COMP	 '\x25'
#define	  OP_COMPH	'\x26'
#define	  OP_NUM	   '\x27'
#define	  OP_NUMT	'\x28'
#define	  OP_MEM  	'\x29'
#define	  OP_FALSE	'\x30'
#define	  OP_TRUE	'\x31'
#define	  OP_CHRI 	'\x32'
#define	  OP_CHRPLUS	'\x33'
#define	  OP_CHRMIN	'\x34'
#define	  OP_BEEP   	'\x35'
#define	  OP_BREAK	'\x36'
#define	  OP_UPD	    '\x37'
#define	  OP_DTRIM1	'\x39'
#define	  OP_DTRIM2	'\x40'
#define	  OP_DTRIM3	'\x41'
#define	  OP_DTRIM4	'\x42'
#define	 OP_RAL        '\x44'
#define	 OP_MODUL   '\x45'
#define	 OP_SHOWT 	'\x46'
#define	 OP_SHOW	  '\x47'
#define   OP_PLAYT    '\x48'
#define   OP_PLAY      '\x49'
#define   OP_PLAYS    '\x50'
#define   OP_PLAYV    '\x51'
#define   OP_JETI       '\x52'
#define   OP_JETIX       '\x53'
#define   OP_VAR       '\x54'
#define   OP_DELTA      '\x55'
#define	OP_COMPA 	 '\x56'
#define	OP_IFNOT  	 '\x57'
#define   OP_JETIRX       '\x58'
#define   OP_LOG         '\x59'
#define   OP_PID          '\x60'
#define   OP_PIDA         '\x61'
#define   OP_VART         '\x62'
#define   OP_VARF         '\x63'
#define   OP_HOLD         '\x64'
#define   OP_EOF          '\x99'  

#endif
